"""
This package contains tools to turn ActivityPub messages
into something cleaner. Its takes include

- Normalization
- Validation


"""

from pydantic import BaseModel, Field

from .types import Activity, Object, Actor, Collection

from .normalize import normalize_activity, normalize_object
from .actor import normalize_actor
from .collection import normalize_collection


class NormalizationResult(BaseModel):
    object: Object | None
    activity: Activity | None
    embedded_object: Object | None = Field(None, serialization_alias="embeddedObject")
    actor: Actor | None
    collection: Collection | None


def normalize(data: dict):
    def result_or_none(func):
        try:
            return func(data)
        except Exception:
            return None

    activity = result_or_none(normalize_activity)

    if activity and isinstance(activity.object, Object):
        embedded_object = activity.object
        activity.object = embedded_object.id
    else:
        embedded_object = None

    return NormalizationResult(
        object=result_or_none(normalize_object),
        activity=activity,
        embedded_object=embedded_object,
        actor=result_or_none(normalize_actor),
        collection=result_or_none(normalize_collection),
    )



from typing import List, Any
import nh3


def list_from_value(value: Any) -> List[Any] | None:
    """Transforms a list into a value

    ```pycon
    >>> list_from_value(["aaa"])
    ['aaa']

    >>> list_from_value("aaa")
    ['aaa']

    >>> list_from_value({"a": 1})
    [{'a': 1}]

    >>> list_from_value([])

    >>> list_from_value(None)

    ```


    :returns: A list or None in case of an empty list or None as argument

    """

    if isinstance(value, list):
        if len(value) == 0:
            return None
        return value
    if isinstance(value, str) or isinstance(value, dict):
        return [value]

    return None


allowed_html_tags = {
    "a",
    "abbr",
    "acronym",
    "b",
    "blockquote",
    "code",
    "em",
    "i",
    "li",
    "ol",
    "strong",
    "ul",
    "p",
    "br",
    "span",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
}
"""The currently allowed list of html tags"""


def sanitize_html(value: str | None) -> str | None:
    """Cleans html

    ```pycon
    >>> sanitize_html("<p>text</p>")
    '<p>text</p>'

    >>> sanitize_html("<script>alert('xss')</script>")
    ''

    ```
    """
    if isinstance(value, str):
        return nh3.clean(value, tags=allowed_html_tags)
    return None


def remove_html(value: str | None) -> str | None:
    """Removes html

    ```pycon
    >>> remove_html('<a href="http://location.test">location.test</p>')
    'location.test'

    ```
    """

    if isinstance(value, str):
        return nh3.clean(value, tags=set())
    return None



import click
import json
from pydantic import BaseModel

from .types import Actor, Activity, Object


@click.group
def main(): ...


def to_json_schema(model: BaseModel, filename: str):
    schema = model.model_json_schema()

    with open(filename, "w") as f:
        json.dump(schema, f, indent=2)


@main.command()
@click.option("--path", default="docs/schemas/")
def schemas(path):
    to_json_schema(Actor, f"{path}/actor.json")  # type: ignore
    to_json_schema(Activity, f"{path}/activity.json")  # type: ignore
    to_json_schema(Object, f"{path}/object.json")  # type: ignore


if __name__ == "__main__":
    main()



from .types import Actor

from .attachment import normalize_attachment
from .utils import sanitize_html, list_from_value, remove_html

from .types.actor import PropertyValue


def normalize_actor(data: dict) -> Actor | None:
    """Normalizes an ActivityPub actor"""

    actor_type = data.get("type")

    if actor_type not in ["Person", "Service", "Application", "Group"]:
        return None

    avatar = data.get("icon")
    if avatar:
        avatar = normalize_attachment(avatar)

    attachments = list_from_value(data.get("attachment"))
    if attachments:
        attachments = [
            normalize_property_value(attachment) for attachment in attachments
        ]

    return Actor.model_validate(
        {
            "@context": data.get("@context"),
            "type": actor_type,
            "id": data.get("id"),
            "inbox": data.get("inbox"),
            "outbox": data.get("outbox"),
            "icon": avatar,
            "summary": sanitize_html(data.get("summary")),
            "name": data.get("name"),
            "alsoKnownAs": list_from_value(data.get("alsoKnownAs")),
            "attachments": attachments,
        }
    )


def normalize_property_value(data: dict) -> PropertyValue | dict:
    if data.get("type") == "PropertyValue":
        return PropertyValue.model_validate(
            {
                "name": remove_html(data.get("name")),
                "value": remove_html(data.get("value")),
            }
        )
    return data



import pytest

from .content import normalize_content


@pytest.mark.parametrize(
    "input,output",
    [({}, None), ({"content": "moo"}, "moo"), ({"contentMap": {"en": "moo"}}, "moo")],
)
def test_basic_content(input, output):
    result = normalize_content(input)

    assert result == output



"""Routings to normalize an ActivityPub activity"""

import logging

from uuid6 import uuid7
from bovine.activitystreams.utils import id_for_object

from .types import Activity, Object
from .utils import list_from_value, sanitize_html
from .attachment import normalize_attachments
from .content import normalize_content

logger = logging.getLogger(__name__)


def normalize_to(value, actor):
    """Normalizes the to value

    ```pycon
    >>> normalize_to(None, "http://actor.example")
    ['http://actor.example']

    >>> normalize_to("http://to.example", None)
    ['http://to.example']

    >>> normalize_to(["http://alice.example", "http://bob.example"], None)
    ['http://alice.example', 'http://bob.example']

    ```
    """
    if value is None:
        return [actor]
    return list_from_value(value)


def normalize_id(activity):
    """
    Creates a normalized id

    ```pycon
    >>> normalize_id({"id": "http://id.example"})
    'http://id.example'

    >>> normalize_id({})
    Traceback (most recent call last):
        ...
    ValueError: Cannot fake id if actor is not present

    ```
    """
    result = activity.get("id")
    if result is not None:
        return result
    actor_id = id_for_object(activity.get("actor"))

    if actor_id is None:
        raise ValueError("Cannot fake id if actor is not present")

    return f"{actor_id}#fake_id" + str(uuid7())


def normalize_activity(activity: dict, actor: str | None = None) -> Activity:
    """
    Normalizes activities.

    :param activity: The activity being normalized
    :param actor: Actor receiving this activity
    :returns:
    """
    try:
        obj = activity.get("object")
        if isinstance(obj, dict):
            try:
                obj = normalize_object(obj)
            except Exception:
                if isinstance(obj, dict):
                    obj = obj.get("id")

        return Activity.model_validate(
            {
                "@context": activity.get("@context"),
                "id": normalize_id(activity),
                "type": activity.get("type"),
                "actor": id_for_object(activity.get("actor")),
                "object": obj,
                "to": normalize_to(activity.get("to"), actor),
                "cc": list_from_value(activity.get("cc")),
                "published": activity.get("published"),
                "target": activity.get("target"),
                "content": activity.get("content"),
            }
        )
    except Exception as e:
        logger.exception(e)
        logger.info(activity)

        raise e


def normalize_url(url):
    if url is None:
        return
    if isinstance(url, str):
        url = {"type": "Link", "href": url}

    return list_from_value(url)


def normalize_to_id(attributed_to) -> str | None:
    if isinstance(attributed_to, list) and len(attributed_to) == 1:
        attributed_to = attributed_to[0]

    return id_for_object(attributed_to)  # type: ignore


def normalize_object(obj: dict) -> Object:
    """Normalizes an object

    :params obj: The object to be normalized
    :returns:
    """
    return Object.model_validate(
        {
            "@context": obj.get("@context"),
            "id": obj.get("id"),
            "type": obj.get("type"),
            "attributedTo": normalize_to_id(obj.get("attributedTo")),
            "obj": obj.get("obj"),
            "to": list_from_value(obj.get("to")),
            "cc": list_from_value(obj.get("cc")),
            "published": obj.get("published"),
            "target": obj.get("target"),
            "content": normalize_content(obj),
            "attachment": normalize_attachments(obj.get("attachment")),
            "sensitive": obj.get("sensitive"),
            "summary": sanitize_html(obj.get("summary")),
            "tag": list_from_value(obj.get("tag")),
            "url": normalize_url(obj.get("url")),
            "inReplyTo": normalize_to_id(obj.get("inReplyTo")),
        }
    )



from cattle_grid.extensions import Extension

from . import normalize

extension = Extension(name="muck out", module=__name__)


@extension.transform(inputs=["raw"], outputs=["parsed"])
async def muck_out(data: dict):
    return {"parsed": normalize(data["raw"])}



import pytest
from pydantic import ValidationError

from .types import Activity, Object
from .normalize import normalize_activity, normalize_object


@pytest.mark.parametrize("activity", [{}])
def test_normalize_activity_invalid(activity):
    with pytest.raises(ValueError):
        normalize_activity(activity)


@pytest.mark.parametrize("obj", [{}])
def test_normalize_object_invalid(obj):
    with pytest.raises(ValidationError):
        normalize_object(obj)


def activity_with_default_context(**kwargs):
    return Activity.model_validate(
        {"@context": "https://www.w3.org/ns/activitystreams", **kwargs}
    )


@pytest.mark.parametrize(
    ["activity", "expected"],
    [
        (
            {
                "@context": "https://www.w3.org/ns/activitystreams",
                "id": "http://remote.test/activity_id",
                "type": "Activity",
                "actor": "http://remote.test/actor",
                "to": "http://actor.example",
                "object": "http://actor.example/object",
            },
            activity_with_default_context(
                id="http://remote.test/activity_id",
                type="Activity",
                actor="http://remote.test/actor",
                to=["http://actor.example"],
                object="http://actor.example/object",
            ),
        ),
        (
            {
                "@context": "https://www.w3.org/ns/activitystreams",
                "id": "http://remote.test/activity_id",
                "type": "Activity",
                "actor": "http://remote.test/actor",
                "to": "http://actor.example",
                "cc": "http://other.example",
                "object": "http://actor.example/object",
                "published": "2024-09-26T18:35:42Z",
                "target": "http://actor.example/target",
                "content": "moo",
            },
            activity_with_default_context(
                id="http://remote.test/activity_id",
                type="Activity",
                actor="http://remote.test/actor",
                to=["http://actor.example"],
                cc=["http://other.example"],
                object="http://actor.example/object",
                published="2024-09-26T18:35:42Z",
                target="http://actor.example/target",
                content="moo",
            ),
        ),
        (
            {
                "@context": "https://www.w3.org/ns/activitystreams",
                "id": "http://remote.test/activity_id",
                "type": "Activity",
                "actor": {"type": "Person", "id": "http://remote.test/actor"},
                "to": "http://actor.example",
                "object": "http://actor.example/object",
            },
            activity_with_default_context(
                id="http://remote.test/activity_id",
                type="Activity",
                actor="http://remote.test/actor",
                to=["http://actor.example"],
                object="http://actor.example/object",
            ),
        ),
    ],
)
def test_normalize_activity_valid(activity, expected):
    result = normalize_activity(activity)

    assert result == expected


@pytest.mark.parametrize(
    ["new_values", "new_expected"],
    [
        ({}, {}),
        (
            {"attachment": {"href": "https://funfedi.dev", "type": "Link"}},
            {"attachment": [{"href": "https://funfedi.dev", "type": "Link"}]},
        ),
        ({"sensitive": True}, {"sensitive": True}),
        ({"content": "<b>bold</b>"}, {"content": "<b>bold</b>"}),
        ({"summary": "test"}, {"summary": "test"}),
        (
            {"summary": "test <script>alert('hi')</script>"},
            {"summary": "test "},
        ),
        (
            {"tag": {"type": "Hashtag", "name": "#test"}},
            {"tag": [{"type": "Hashtag", "name": "#test"}]},
        ),
        (
            {"url": "https://remote.test/"},
            {"url": [{"type": "Link", "href": "https://remote.test/"}]},
        ),
        (
            {"inReplyTo": {"id": "https://remote.test/to_reply", "type": "Note"}},
            {"inReplyTo": "https://remote.test/to_reply"},
        ),
    ],
)
def test_normalize_object(new_values, new_expected):
    obj = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/object_id",
        "type": "Activity",
        "attributedTo": "http://remote.test/actor",
        "to": "http://actor.example",
        "content": "text",
        **new_values,
    }
    expected = Object.model_validate(
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "id": "http://remote.test/object_id",
            "type": "Activity",
            "attributedTo": "http://remote.test/actor",
            "to": ["http://actor.example"],
            "content": "text",
            **new_expected,
        }
    )

    result = normalize_object(obj)

    assert result == expected


@pytest.mark.parametrize(
    "example",
    [
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "id": "http://mastodon/users/hippo#rejects/follows/1",
            "type": "Reject",
            "actor": "http://mastodon/users/hippo",
            "object": {
                "id": "http://abel/actor/Ahua3T7U2umku3O4U3wrqg#1.8q4hyllt5j",
                "type": "Follow",
                "actor": "http://abel/actor/Ahua3T7U2umku3O4U3wrqg",
                "object": "http://mastodon/users/hippo",
            },
        },
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "id": "http://mastodon/users/hippo#accepts/follows/",
            "type": "Accept",
            "actor": "http://mastodon/users/hippo",
            "object": {
                "id": "http://abel/actor/lTS2tmsmXuLqcFBWYBjnbQ#1.8huk5ukb1i",
                "type": "Follow",
                "actor": "http://abel/actor/lTS2tmsmXuLqcFBWYBjnbQ",
                "object": "http://mastodon/users/hippo",
            },
        },
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Accept",
            "actor": "http://abel/actor/u4Kq-wez2koDZsLQfZKN5g",
            "to": ["http://banach/actor/T27aBR0fkpFwEkgzVTlFdg"],
            "published": "2024-11-07T14:50:28Z",
            "object": "follow:405205da-1723-4099-a69f-6d63b4937b34",
        },
    ],
)
def test_normalize_activity_examples(example):
    normalize_activity(example, actor="http://abel/actor/Ahua3T7U2umku3O4U3wrqg")



import re
import pytest

from .actor import normalize_actor, normalize_property_value
from .types import Actor
from .types.actor import PropertyValue


def test_normalize_actor_failure():
    assert normalize_actor({}) is None


actor_base = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Person",
    "id": "https://example.com/users/alice",
}


@pytest.mark.parametrize(
    "data",
    [
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Person",
            "id": "https://example.com/users/alice",
            "inbox": "https://example.com/users/alice/inbox",
        },
        {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Person",
            "id": "https://example.com/users/alice",
            "inbox": "https://example.com/users/alice/inbox",
            "icon": {
                "type": "Image",
                "url": "https://example.com/users/alice/icon.png",
            },
        },
    ],
)
def test_normalize_actor(data):
    assert isinstance(normalize_actor(data), Actor)


def to_snake_case(string):
    return re.sub("([A-Z]+)", r"_\1", string).lower()


@pytest.mark.parametrize(
    "field, value, mapped",
    [
        ("summary", "<p>Hello, world!</p>", "<p>Hello, world!</p>"),
        (
            "summary",
            "<p>Test<script>alert('hi')</script></p>",
            "<p>Test</p>",
        ),
        ("name", "Alice", "Alice"),
        (
            "alsoKnownAs",
            ["https://example.com/users/alice"],
            ["https://example.com/users/alice"],
        ),
    ],
)
def test_normalize_actor_field(field, value, mapped):
    data = actor_base.copy()
    data[field] = value
    actor = normalize_actor(data)
    assert getattr(actor, to_snake_case(field)) == mapped


def test_normalize_property_value_dict():
    data = {"type": "Other", "name": "wooo"}

    assert normalize_property_value(data) == data


def test_normalize_property_value():
    data = {
        "type": "PropertyValue",
        "name": "Blog",
        "value": '<a href="http://value.example/" target="_blank" rel="nofollow noopener noreferrer me" translate="no"><span class="invisible">http://</span><span class="">value.example/</span><span class="invisible"></span></a>',
    }

    result = normalize_property_value(data)

    assert isinstance(result, PropertyValue)

    assert result.value == "http://value.example/"



from click.testing import CliRunner

from .__main__ import main


def test_schemas(tmp_path):
    runner = CliRunner()

    runner.invoke(main, ["schemas", "--path", tmp_path])

    assert len(list(tmp_path.glob("*.json"))) == 3



from .utils import sanitize_html


def normalize_content(data: dict) -> str | None:
    """Normalizes the content of an ActivityPub object"""
    result = data.get("content")
    if isinstance(result, str):
        return sanitize_html(result)

    content_map = data.get("contentMap")

    if isinstance(content_map, dict) and len(content_map) > 0:
        return sanitize_html(list(content_map.values())[0])

    return None



from typing import List
from bovine.activitystreams.utils import id_for_object

from .types import Collection


def convert_items(items: List[dict | str]) -> List[str]:
    """Converts a list of items to a list of ids

    ```pycon
    >>> convert_items([{"id": "https://site.example/1"}, "https://site.example/2"])
    ['https://site.example/1', 'https://site.example/2']

    ```
    """
    return [x for x in (id_for_object(item) for item in items) if x]


def normalize_collection(collection: dict) -> Collection | None:
    """Normalizes a collection"""

    type = collection.get("type")
    if type is None or type not in (
        "Collection",
        "OrderedCollection",
        "CollectionPage",
        "OrderedCollectionPage",
    ):
        return None
    count = collection.get("totalItems")
    items = collection.get("orderedItems")
    if items is None:
        items = collection.get("items")

    data = {
        "id": collection.get("id"),
        "type": type,
        "count": count,
        "first": collection.get("first"),
        "last": collection.get("last"),
        "next": collection.get("next"),
        "prev": collection.get("prev"),
        "items": items,
    }

    if isinstance(data.get("first"), dict):
        data["next"] = data["first"].get("next")
        data["items"] = data["first"].get("items")
        data["first"] = None

    if data["items"]:
        data["items"] = convert_items(data["items"])

    return Collection.model_validate(data)



from pydantic import BaseModel, Field


class PropertyValue(BaseModel):
    """
    Key value pairs in the attachment of an actor
    as used by Mastodon
    """

    type: str = Field("PropertyValue", description="""Fixed type for serialization""")

    name: str = Field(
        examples=["Pronouns"],
        description="""
    Key of the value
    """,
    )

    value: str = Field(
        examples=["They/them"],
        description="""
    Value
    """,
    )



from .collection import normalize_collection
from .types import Collection


def test_collection_invalid():
    assert normalize_collection({}) is None


def test_collection_mastodon_outbox_type():
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://mastodon.example/users/alice/outbox",
        "type": "OrderedCollection",
        "totalItems": 12360,
        "first": "http://mastodon.example/users/alice/outbox?page=true",
        "last": "http://mastodon.example/users/alice/outbox?min_id=0&page=true",
    }

    result = normalize_collection(data)

    assert isinstance(result, Collection)

    assert result.id == data["id"]
    assert result.count == data["totalItems"]
    assert result.first == data["first"]
    assert result.last == data["last"]


def test_collection_mastodon_outbox_page_type():
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://mastodon.example/users/alice/outbox?page=true",
        "type": "OrderedCollectionPage",
        "next": "http://mastodon.example/users/alice/outbox?max_id=12&page=true",
        "prev": "http://mastodon.example/users/alice/outbox?min_id=17&page=true",
        "orderedItems": ["item1", {"aaa": "bbb"}, {"id": "item2"}],
    }

    result = normalize_collection(data)

    assert isinstance(result, Collection)

    assert result.id == data["id"]
    assert result.next == data["next"]
    assert result.prev == data["prev"]

    assert result.items == ["item1", "item2"]


def test_collection_mastodon_likes_type():
    data = {
        "id": "http://mastodon.example/users/alice/statuses/2343242/likes",
        "type": "Collection",
        "totalItems": 0,
    }

    result = normalize_collection(data)

    assert isinstance(result, Collection)
    assert result.count == 0


def test_collection_mastodon_replies_type():
    data = {
        "id": "http://mastodon.example/users/alice/statuses/2343242/replies",
        "type": "Collection",
        "first": {
            "type": "CollectionPage",
            "next": "http://mastodon.example/users/alice/statuses/2343242/replies?only_other_accounts=true&page=true",
            "partOf": "http://mastodon.example/users/alice/statuses/2343242/replies",
            "items": [],
        },
    }

    result = normalize_collection(data)

    assert isinstance(result, Collection)


def test_collection_page():
    data = {
        "id": "http://mastodon.example/users/alice/statuses/2343242/replies?fake=true",
        "type": "CollectionPage",
        "next": "http://mastodon.example/users/alice/statuses/2343242/replies?only_other_accounts=true&page=true",
        "partOf": "http://mastodon.example/users/alice/statuses/2343242/replies",
        "items": ["http://something.example"],
    }

    result = normalize_collection(data)

    assert isinstance(result, Collection)
    assert result.items == data["items"]



from pydantic import Field
from typing import List, Any, Dict

from .actor import PropertyValue
from .common import Common, CommonAll


class Object(Common):
    attributedTo: str = Field(
        examples=["https://actor.example/"],
        description="""
    id of the actor that authored this object
    """,
    )
    content: str = Field(
        description="""
    The content of the object
    """
    )
    summary: str | None = Field(
        None,
        description="""
    The summary of the object
    """,
    )
    name: str | None = Field(
        None,
        description="""
    The name of the object
    """,
    )
    attachment: List[Dict[str, Any]] | None = Field(
        None,
        description="""
    A list of objects that are attached to the original object
    """,
    )
    tag: List[Dict[str, Any]] | None = Field(
        None,
        description="""
    A list of objects that expand on the content of the object
    """,
    )
    url: List[Dict[str, Any] | str] | None = Field(
        None,
        description="""
    A list of urls that expand on the content of the object
    """,
    )
    sensitive: bool | None = Field(
        None,
        description="""
    Marks the object as sensitive. Currently, used by everyone, a better way would be an element of the tag list that labels the object as sensitive due a reason
    """,
    )
    inReplyTo: str | None = Field(
        None,
        description="""
    The object being replied to. Currently a string. Not sure if this is what I want.
    """,
    )
    id: str = Field(
        ...,
        examples=["https://actor.example/some_id"],
        description="""
    id of the activity, can be assumed to be globally unique. Some activities such as a Follow request will require an id to be valid. Servers may assume an id to be required. As assigning an id is 'trivial', one should assign one.
    """,
    )
    type: str = Field(
        ...,
        examples=["Follow", "Accept", "Create", "Undo", "Like"],
        description="""
    Type of the activity. Side effects of this activity are determine by this type.
    """,
    )
    to: List[str] = Field(
        ...,
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        min_length=1,
        description="""
    Array of actors this activity is addressed to. It is sane to assume that an activity is addressed to at least one person.
    """,
    )


class Activity(Common):
    """
    This represents a first draft of a json-schema that every activities exchanged between servers MUST satisfy and be able to parse. Here 'being able to parse' means making it to the point, where depending on the type, you decide what side effects to perform.

    Generally, the fields actor, to, and cc (and maybe bcc --- not transported) represent how the message is being delivered. The fields actor, type, object, target, content represent how the message is processed by the server.
    """

    actor: str = Field(
        ...,
        examples=["https://actor.example/"],
        description="""
    id of the actor performing this activity. One can assume that the activity is signed by this actor (in some form).
    """,
    )
    object: str | Object | None = Field(None)
    target: str | Dict[str, Any] | None = Field(
        None,
        examples=[
            "https://other.example/target_id",
            {"type": "Note", "content": "meow"},
        ],
        description="""
    The target, not sure if needed, included for completeness
    """,
    )
    content: str | None = Field(
        None,
        examples=["🐮", "❤️"],
        description="""
    The content used for example to represent the Emote for a like
    """,
    )
    id: str = Field(
        ...,
        examples=["https://actor.example/some_id"],
        description="""
    id of the activity, can be assumed to be globally unique. Some activities such as a Follow request will require an id to be valid. Servers may assume an id to be required. As assigning an id is 'trivial', one should assign one.
    """,
    )
    type: str = Field(
        ...,
        examples=["Follow", "Accept", "Create", "Undo", "Like"],
        description="""
    Type of the activity. Side effects of this activity are determine by this type.
    """,
    )
    to: List[str] = Field(
        ...,
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        min_length=1,
        description="""
    Array of actors this activity is addressed to. It is sane to assume that an activity is addressed to at least one person.
    """,
    )


class Actor(CommonAll):
    """Describes an ActivityPub actor"""

    type: str = Field(
        examples=["Person", "Service", "Application"],
        description="""The type of Actor""",
    )

    inbox: str | None = Field(
        None,
        examples=["https://actor.example/inbox"],
        description="""
    The inbox of the actor
    """,
    )

    outbox: str | None = Field(
        None,
        examples=["https://actor.example/outbox"],
        description="""
    The outbox of the actor
    """,
    )

    icon: dict | None = Field(
        None,
        examples=[{"type": "Image", "url": "https://actor.example/icon.png"}],
        description="""
    The avatar of the actor
    """,
    )

    summary: str | None = Field(
        None,
        examples=["My Fediverse account"],
        description="""
    Description of the actor
    """,
    )

    name: str | None = Field(
        None,
        examples=["Alice"],
        description="""
    Display name of the actor
    """,
    )

    also_known_as: List[str] | None = Field(
        None,
        examples=[["https://alice.example", "https://alice.example/profile"]],
        alias="alsoKnownAs",
        description="""
    Other uris associated with the actor
    """,
    )

    attachments: List[dict | PropertyValue] | None = Field(
        None, description="""attachments ... currently used for property values"""
    )


class Collection(CommonAll):
    """Abstracts all the ActivityPub collection concepts"""

    type: str = Field(
        examples=[
            "Collection",
            "OrderedCollection",
            "CollectionPage",
            "OrdererCollectionPage",
        ],
        description="""Type of object""",
    )

    items: List[str] | None = Field(None, description="""The items""")

    next: str | None = Field(None)
    prev: str | None = Field(None)
    first: str | None = Field(None)
    last: str | None = Field(None)
    count: int | None = Field(None)



from pydantic import BaseModel, Field
from typing import List, Any


class CommonAll(BaseModel):
    field_context: str | List[Any] | None = Field(
        None,
        alias="@context",
        examples=[
            "https://www.w3.org/ns/activitystreams",
            ["https://www.w3.org/ns/activitystreams", {"Hashtag": "as:Hashtag"}],
        ],
    )
    id: str = Field(
        ...,
        examples=["https://actor.example/some_id"],
        description="""
    id of the activity or object, can be assumed to be globally unique. Some activities such as a Follow request will require an id to be valid. Servers may assume an id to be required. As assigning an id is 'trivial', one should assign one.
    """,
    )


class Common(CommonAll):
    to: List[str] = Field(
        ...,
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        min_length=1,
        description="""
    Array of actors this activity or object is addressed to. It is sane to assume that an activity is addressed to at least one person.
    """,
    )
    cc: List[str] | None = Field(
        None,
        examples=[
            ["https://bob.example"],
            ["https://alice.example", "https://bob.example"],
        ],
        description="""
    Array of actors this activity or object is carbon copied to.
    """,
    )
    published: str | None = Field(
        None,
        description="""
    Moment of this activity or object being published
    """,
    )
    type: str = Field(
        ...,
        examples=["Follow", "Accept", "Create", "Undo", "Like", "Note"],
        description="""
    Type of the activity or activity. Side effects of this activity are determine by this type.
    """,
    )



import pytest
import json

from zipfile import ZipFile

from . import normalize


def object_data():
    with ZipFile("test_data/samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("object/"):
                yield file_name, json.loads(zip_file.read(file_name))


def activity_data():
    with ZipFile("test_data/samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("activity/"):
                yield file_name, json.loads(zip_file.read(file_name))


failing_objects = [
    "object/object_types_12.json",
    "object/object_types_13.json",
    "object/necessary_properties_1.json",
    "object/necessary_properties_3.json",
    "object/necessary_properties_4.json",
    "object/in_reply_to_4.json",
    "object/attributed_to_2.json",
]


@pytest.mark.parametrize("file_name, data", object_data())
def test_objects(file_name, data):
    normalized = normalize(data)

    if file_name in failing_objects:
        assert normalized.object is None
    else:
        assert normalized.object


@pytest.mark.parametrize("file_name, data", activity_data())
def test_activities(file_name, data):
    normalized = normalize(data)

    assert normalized.activity


def actor_data():
    with ZipFile("test_data/funfedi-samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("actors/"):
                yield file_name, json.loads(zip_file.read(file_name))


@pytest.mark.parametrize("file_name, data", actor_data())
def test_actors(file_name, data):
    normalized = normalize(data)

    assert normalized.actor
